#include "guecs/sfml/backend.hpp"
#include "guecs/sfml/components.hpp"
#include "guecs/ui.hpp"
#include <fmt/xchar.h>
#include <deque>
#include <iostream>

constexpr const int SCREEN_WIDTH=1280;
constexpr const int SCREEN_HEIGHT=720;
constexpr const int FRAME_LIMIT=60;
constexpr const bool VSYNC=true;
bool GO_TO_WORK = false;

using std::string, std::wstring;

enum class Event {
  CLICKER, A_BUTTON
};

struct Shake {
  float scale_factor = 0.05f;
  int frames = 10;
  float ease_rate = 0.1f;
  bool playing = false;
  int current = 0;
  float x=0.0;
  float y=0.0;
  float w=0.0;
  float h=0.0;
  sf::Vector2f initial_scale;

  float ease() {
    float tick = float(frames) / float(current) * ease_rate;
    return (std::sin(tick) + 1.0) / 2.0;
  }

  void init(lel::Cell& cell) {
    x = cell.x;
    y = cell.y;
    w = cell.w;
    h = cell.h;
  }

  void play(guecs::Sprite& sprite) {
    if(!playing) {
      playing = true;
      current = 0;
      initial_scale = sprite.sprite->getScale();
    }
  }

  void render(guecs::Sprite& sprite) {
    current++;

    if(playing && current < frames) {
      float tick = ease();
      sf::Vector2f scale{
        std::lerp(initial_scale.x, initial_scale.x + scale_factor, tick),
          std::lerp(initial_scale.y, initial_scale.y + scale_factor, tick)};

      sprite.sprite->setScale(scale);
    } else {
      playing = false;
      current = 0;
      sprite.sprite->setScale(initial_scale);
    }
  }
};

struct ClickerUI {
  guecs::UI $gui;
  guecs::Entity $clicker;

  ClickerUI() {
    $gui.position(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
    $gui.layout(
        "[FoodBowl|_|*%(300,400)clicker|_|_|_]"
        "[WaterBowl|_|_               |_|_|_]"
        "[Mood|_|_               |_|_|_]"
        "[Money|_|_               |_|_|_]"
        "[Treat|Food|Water|Pet|Work]");
  }

  void init() {
    $gui.set<guecs::Background>($gui.MAIN, {$gui.$parser, {0, 0, 0, 255}});

    for(auto& [name, cell] : $gui.cells()) {
      auto id = $gui.entity(name);
      if(name != "clicker") {
        $gui.set<guecs::Rectangle>(id, {});
        $gui.set<guecs::Effect>(id, {});
        $gui.set<guecs::Label>(id, { guecs::to_wstring(name) });

        if(name == "Work") {
          $gui.set<guecs::Clickable>(id, { [&](auto, auto) {
              GO_TO_WORK = true;
              fmt::println("Going to Work!");
          } });
        } else {
          $gui.set<guecs::Clickable>(id, {
              [&](auto, auto) { handle_button(Event::A_BUTTON); }
          });
        }
      }
    }

    $clicker = $gui.entity("clicker");
    $gui.set<guecs::Sprite>($clicker, {"clicker_the_dog"});
    $gui.set<guecs::Sound>($clicker, {"clicker_bark"});
    $gui.set<guecs::Clickable>($clicker, {
        [&](auto, auto) { handle_button(Event::CLICKER); }
        });

    // custom components need to be initialized manually
    $gui.set_init<Shake>($clicker, {});

    $gui.init();
  }

  void render(sf::RenderWindow& window) {
    auto& shaker = $gui.get<Shake>($clicker);

    if(shaker.playing) {
      auto& sprite = $gui.get<guecs::Sprite>($clicker);
      shaker.render(sprite);
      window.clear();
    }

    $gui.render(window);

    // $gui.debug_layout(window);
  }

  void mouse(float x, float y, bool hover) {
    $gui.mouse(x, y, hover);
  }

  void handle_button(Event ev) {
    using enum Event;
    switch(ev) {
      case CLICKER: {
                      auto& shaker = $gui.get<Shake>($clicker);
                      auto& sprite = $gui.get<guecs::Sprite>($clicker);
                      shaker.play(sprite);
                      fmt::println("CLICKER LOVES YOU!");
                    } break;
      case A_BUTTON:
                    fmt::println("a button clicked");
                    break;

      default:
                    assert(false && "invalid event");
    }
  }
};


struct WorkComputerUI {
  guecs::UI $gui;

  WorkComputerUI() {
    $gui.position(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
    $gui.layout("[computer]");
  }

  void init() {
    $gui.set<guecs::Background>($gui.MAIN, {$gui.$parser});

    auto computer = $gui.entity("computer");
    $gui.set<guecs::Label>(computer, {L"Work Computer"});
    $gui.set<guecs::Sprite>(computer, {"work_computer"});
    $gui.set<guecs::Clickable>(computer, { [&](auto, auto) {
      GO_TO_WORK = false;
      fmt::println("Leaving Work!");
    } });

    $gui.init();
  }

  void render(sf::RenderWindow& window) {
    $gui.render(window);
  }

  void mouse(float x, float y, bool hover) {
    $gui.mouse(x, y, hover);
  }
};


int main() {
  sfml::Backend backend;
  guecs::init(&backend);

  sf::RenderWindow window(sf::VideoMode({SCREEN_WIDTH, SCREEN_HEIGHT}), "Clicker the Dog");
  window.setFramerateLimit(FRAME_LIMIT);
  window.setVerticalSyncEnabled(VSYNC);

  ClickerUI clicker;
  clicker.init();

  WorkComputerUI computer;
  computer.init();

  while(window.isOpen()) {
    while (const auto event = window.pollEvent()) {
      if(event->is<sf::Event::Closed>()) {
        window.close();
      }

      if(const auto* mouse = event->getIf<sf::Event::MouseButtonPressed>()) {
        if(mouse->button == sf::Mouse::Button::Left) {
          sf::Vector2f pos = window.mapPixelToCoords(mouse->position);
          if(GO_TO_WORK) {
            computer.mouse(pos.x, pos.y, false);
          } else {
            clicker.mouse(pos.x, pos.y, false);
          }
        }
      }
    }

    if(GO_TO_WORK) {
      window.clear();
      computer.render(window);
    } else {
      clicker.render(window);
    }
    window.display();
  }
}
